@@ -19,13 +19,12 @@ from __future__ import unicode_literals
 
 __intname__ = "command_runner"
 __author__ = "Orsiris de Jong"
-__copyright__ = "Copyright (C) 2015-2024 Orsiris de Jong for NetInvent SASU"
+__copyright__ = "Copyright (C) 2015-2025 Orsiris de Jong for NetInvent"
 __licence__ = "BSD 3 Clause"
-__version__ = "1.7.0"
-__build__ = "2024091501"
+__version__ = "1.7.5"
+__build__ = "2025090901"
 __compat__ = "python2.7+"
 
-import io
 import os
 import shlex
 import subprocess
@@ -35,32 +34,90 @@ from logging import getLogger
 from time import sleep
 
 
+# Avoid checking os type numerous times
+os_name = os.name
+
+
+# Don't bother with an ImportError since we need command_runner to work without dependencies
 try:
     import psutil
-except ImportError:
-    # Don't bother with an error since we need command_runner to work without dependencies
-    pass
-try:
+
     # Also make sure we directly import priority classes so we can reuse them
-    if os.name == "nt":
+    if os_name == "nt":
         from psutil import (
-            ABOVE_NORMAL_PRIORITY_CLASS,
+            # ABOVE_NORMAL_PRIORITY_CLASS,
             BELOW_NORMAL_PRIORITY_CLASS,
             HIGH_PRIORITY_CLASS,
             IDLE_PRIORITY_CLASS,
             NORMAL_PRIORITY_CLASS,
             REALTIME_PRIORITY_CLASS,
         )
-        from psutil import IOPRIO_HIGH, IOPRIO_NORMAL, IOPRIO_LOW, IOPRIO_VERYLOW
+        from psutil import (
+            IOPRIO_HIGH,
+            IOPRIO_NORMAL,
+            IOPRIO_LOW,
+            # IOPRIO_VERYLOW,
+        )
     else:
         from psutil import (
             IOPRIO_CLASS_BE,
             IOPRIO_CLASS_IDLE,
-            IOPRIO_CLASS_NONE,
+            # IOPRIO_CLASS_NONE,
             IOPRIO_CLASS_RT,
         )
 except (ImportError, AttributeError):
-    pass
+    if os_name == "nt":
+        BELOW_NORMAL_PRIORITY_CLASS = 16384
+        HIGH_PRIORITY_CLASS = 128
+        NORMAL_PRIORITY_CLASS = 32
+        REALTIME_PRIORITY_CLASS = 256
+        IDLE_PRIORITY_CLASS = 64
+        IOPRIO_HIGH = 3
+        IOPRIO_NORMAL = 2
+        IOPRIO_LOW = 1
+    else:
+        IOPRIO_CLASS_IDLE = 3
+        IOPRIO_CLASS_BE = 2
+        IOPRIO_CLASS_RT = 1
+
+
+# Python 2.7 does not have priorities defined in subprocess module, but psutil has
+# Since Windows and Linux use different possible values, let's simplify things by
+# allowing 5 process priorities: verylow (idle), low, normal, high, rt
+# and 3 process io priorities: low, normal, high
+# For IO, rt == high
+if os_name == "nt":
+    PRIORITIES = {
+        "process": {
+            "verylow": IDLE_PRIORITY_CLASS,
+            "low": BELOW_NORMAL_PRIORITY_CLASS,
+            "normal": NORMAL_PRIORITY_CLASS,
+            "high": HIGH_PRIORITY_CLASS,
+            "rt": REALTIME_PRIORITY_CLASS,
+        },
+        "io": {
+            "low": IOPRIO_LOW,
+            "normal": IOPRIO_NORMAL,
+            "high": IOPRIO_HIGH,
+        },
+    }
+else:
+    PRIORITIES = {
+        "process": {
+            "verylow": 20,
+            "low": 15,
+            "normal": 0,
+            "high": -15,
+            "rt": -20,
+        },
+        "io": {
+            "low": IOPRIO_CLASS_IDLE,
+            "normal": IOPRIO_CLASS_BE,
+            "high": IOPRIO_CLASS_RT,
+        },
+    }
+
+
 try:
     import signal
 except ImportError:
@@ -75,7 +132,7 @@ import threading
 
 # Python 2.7 compat fixes (missing typing)
 try:
-    from typing import Union, Optional, List, Tuple, NoReturn, Any, Callable
+    from typing import Union, Optional, List, Tuple, Any, Callable
 except ImportError:
     pass
 
@@ -122,7 +179,11 @@ except AttributeError:
             self.timeout = timeout
             self.output = output
             self.stderr = stderr
-            super().__init__(*args, **kwargs)
+            try:
+                super().__init__(*args, **kwargs)
+            except TypeError:
+                # python 2.7 needs super(Baseclass, self)
+                super(BaseException, self).__init__(*args, **kwargs)
 
         def __str__(self):
             return "Command '%s' timed out after %s seconds" % (self.cmd, self.timeout)
@@ -145,7 +206,11 @@ class InterruptGetOutput(BaseException):
 
     def __init__(self, output, *args, **kwargs):
         self._output = output
-        super().__init__(*args, **kwargs)
+        try:
+            super().__init__(*args, **kwargs)
+        except TypeError:
+            # python 2.7 needs super(Baseclass, self)
+            super(BaseException, self).__init__(*args, **kwargs)
 
     @property
     def output(self):
@@ -159,7 +224,11 @@ class KbdInterruptGetOutput(InterruptGetOutput):
 
     def __init__(self, output, *args, **kwargs):
         self._output = output
-        super().__init__(output, *args, **kwargs)
+        try:
+            super().__init__(*args, **kwargs)
+        except TypeError:
+            # python 2.7 needs super(Baseclass, self)
+            super(InterruptGetOutput, self).__init__(*args, **kwargs)
 
     @property
     def output(self):
@@ -173,7 +242,11 @@ class StopOnInterrupt(InterruptGetOutput):
 
     def __init__(self, output, *args, **kwargs):
         self._output = output
-        super().__init__(output, *args, **kwargs)
+        try:
+            super().__init__(*args, **kwargs)
+        except TypeError:
+            # python 2.7 needs super(Baseclass, self)
+            super(InterruptGetOutput, self).__init__(*args, **kwargs)
 
     @property
     def output(self):
@@ -238,56 +311,56 @@ logger = getLogger(__intname__)
 PIPE = subprocess.PIPE
 
 
+def _validate_process_priority(
+    priority,  # type: Union[int, str]
+):
+    # type: (...) -> int
+    """
+    Check if priority int is valid
+    """
+
+    def _raise_prio_error(priority, reason):
+        raise ValueError(
+            "Priority not valid ({}): {}. Please use one of {}".format(
+                reason, priority, ", ".join(list(PRIORITIES["process"].keys()))
+            )
+        )
+
+    if isinstance(priority, int):
+        if os_name == "nt":
+            _raise_prio_error(priority, "windows does not accept ints as priority")
+        if -20 <= priority <= 20:
+            _raise_prio_error(priority, "priority out of range")
+    elif isinstance(priority, str):
+        try:
+            priority = PRIORITIES["process"][priority.lower()]
+        except KeyError:
+            _raise_prio_error(priority, "priority does not exist")
+    return priority
+
+
 def _set_priority(
     pid,  # type: int
     priority,  # type: Union[int, str]
     priority_type,  # type: str
 ):
     """
-    Set process and / or io priorities
-    Since Windows and Linux use different possible values, let's simplify things by allowing 3 prioriy types
+    Set process and / or io prioritie
     """
     priority = priority.lower()
 
     if priority_type == "process":
-        if isinstance(priority, int) and os.name != "nt" and -20 <= priority <= 20:
-            raise ValueError("Bogus process priority int given: {}".format(priority))
-        if priority not in ["low", "normal", "high"]:
-            raise ValueError(
-                "Bogus {} priority given: {}".format(priority_type, priority)
-            )
-
-    if priority_type == "io" and priority not in ["low", "normal", "high"]:
-        raise ValueError("Bogus {} priority given: {}".format(priority_type, priority))
-
-    if os.name == "nt":
-        priorities = {
-            "process": {
-                "low": BELOW_NORMAL_PRIORITY_CLASS,
-                "normal": NORMAL_PRIORITY_CLASS,
-                "high": HIGH_PRIORITY_CLASS,
-            },
-            "io": {"low": IOPRIO_LOW, "normal": IOPRIO_NORMAL, "high": IOPRIO_HIGH},
-        }
-    else:
-        priorities = {
-            "process": {"low": 15, "normal": 0, "high": -15},
-            "io": {
-                "low": IOPRIO_CLASS_IDLE,
-                "normal": IOPRIO_CLASS_BE,
-                "high": IOPRIO_CLASS_RT,
-            },
-        }
-
-    if priority_type == "process":
-        # Allow direct priority nice settings under linux
-        if isinstance(priority, int):
-            _priority = priority
-        else:
-            _priority = priorities[priority_type][priority]
+        _priority = _validate_process_priority(priority)
         psutil.Process(pid).nice(_priority)
     elif priority_type == "io":
-        psutil.Process(pid).ionice(priorities[priority_type][priority])
+        valid_io_priorities = list(PRIORITIES["io"].keys())
+        if priority not in valid_io_priorities:
+            raise ValueError(
+                "Bogus {} priority given: {}. Please use one of {}".format(
+                    priority_type, priority, ", ".join(valid_io_priorities)
+                )
+            )
+        psutil.Process(pid).ionice(PRIORITIES[priority_type][priority])
     else:
         raise ValueError("Bogus priority type given.")
 
@@ -320,7 +393,7 @@ def to_encoding(
     # type: (...) -> str
     """
     Convert bytes output to string and handles conversion errors
-    Varation of ofunctions.string_handling.safe_string_convert
+    Variation of ofunctions.string_handling.safe_string_convert
     """
 
     if not encoding:
@@ -352,7 +425,7 @@ def kill_childs_mod(
     """
     Inline version of ofunctions.kill_childs that has no hard dependency on psutil
 
-    Kills all childs of pid (current pid can be obtained with os.getpid())
+    Kills all children of pid (current pid can be obtained with os.getpid())
     If no pid given current pid is taken
     Good idea when using multiprocessing, is to call with atexit.register(ofunctions.kill_childs, os.getpid(),)
 
@@ -436,7 +509,7 @@ def kill_childs_mod(
                     pid, 15
                 )  # 15 being signal.SIGTERM or SIGKILL depending on the platform
             except OSError as exc:
-                if os.name == "nt":
+                if os_name == "nt":
                     # We'll do an ugly hack since os.kill() has some pretty big caveats on Windows
                     # especially for Python 2.7 where we can get Access Denied
                     os.system("taskkill /F /pid {}".format(pid))
@@ -520,13 +593,13 @@ def command_runner(
     # Choose default encoding when none set
     # cp437 encoding assures we catch most special characters from cmd.exe
     # Unless encoding=False in which case nothing gets encoded except Exceptions and logger strings for Python 2
-    error_encoding = "cp437" if os.name == "nt" else "utf-8"
+    error_encoding = "cp437" if os_name == "nt" else "utf-8"
     if encoding is None:
         encoding = error_encoding
 
     # Fix when unix command was given as single string
     # This is more secure than setting shell=True
-    if os.name == "posix":
+    if os_name == "posix":
         if not shell and isinstance(command, str):
             command = shlex.split(command)
         elif shell and isinstance(command, list):
@@ -543,7 +616,7 @@ def command_runner(
         windows_no_window
         and sys.version_info[0] >= 3
         and sys.version_info[1] >= 7
-        and os.name == "nt"
+        and os_name == "nt"
     ):
         # Disable the following pylint error since the code also runs on nt platform, but
         # triggers an error on Unix
@@ -625,7 +698,7 @@ def command_runner(
 
         # We also need to check that stream has readline, in case we're writing to files instead of PIPE
 
-        # Another magnificient python 2.7 fix
+        # Another magnificent python 2.7 fix
         # So we need to convert sentinel_char which would be unicode because of unicode_litterals
         # to str which is the output format from stream.readline()
 
@@ -662,6 +735,24 @@ def command_runner(
                 break
             sleep(1)
 
+    def heartbeat_thread(
+        process,  # type: Union[subprocess.Popen[str], subprocess.Popen]
+        heartbeat,  # type: int
+    ):
+        """
+        Just a shorthand to run the heartbeat thread
+        """
+        if heartbeat:
+            heartbeat_thread = threading.Thread(
+                target=_heartbeat_thread,
+                args=(
+                    process,
+                    heartbeat,
+                ),
+            )
+            heartbeat_thread.daemon = True
+            heartbeat_thread.start()
+
     def _poll_process(
         process,  # type: Union[subprocess.Popen[str], subprocess.Popen]
         timeout,  # type: int
@@ -687,7 +778,7 @@ def command_runner(
             # type: (...) -> None
             """
             Simple subfunction to check whether timeout is reached
-            Since we check this alot, we put it into a function
+            Since we check this a lot, we put it into a function
             """
             if timeout and (datetime.now() - begin_time).total_seconds() > timeout:
                 kill_childs_mod(process.pid, itself=True, soft_kill=False)
@@ -699,16 +790,8 @@ def command_runner(
                 raise StopOnInterrupt(_get_error_output(output_stdout, output_stderr))
 
         begin_time = datetime.now()
-        if heartbeat:
-            heartbeat_thread = threading.Thread(
-                target=_heartbeat_thread,
-                args=(
-                    process,
-                    heartbeat,
-                ),
-            )
-            heartbeat_thread.daemon = True
-            heartbeat_thread.start()
+
+        heartbeat_thread(process, heartbeat)
 
         if encoding is False:
             output_stdout = output_stderr = b""
@@ -755,7 +838,11 @@ def command_runner(
                             if stdout_destination == "queue":
                                 stdout.put(line)
                             if live_output:
-                                sys.stdout.write(line)
+                                if encoding == False:
+                                    # We need to allow binary output too, hence using sys.stdout.buffer instead of sys.stdout
+                                    sys.stdout.buffer.write(line)
+                                else:
+                                    sys.stdout.write(line)
                             output_stdout += line
 
                 if stderr_read_queue:
@@ -820,7 +907,7 @@ def command_runner(
                 break
             if process.poll() is not None:
                 break
-            # We definitly need some sleep time here or else we will overload CPU
+            # We definitely need some sleep time here or else we will overload CPU
             sleep(check_interval)
 
     def _monitor_process(
@@ -847,16 +934,8 @@ def command_runner(
         )
         thread.daemon = True  # was setDaemon(True) which has been deprecated
         thread.start()
-        if heartbeat:
-            heartbeat_thread = threading.Thread(
-                target=_heartbeat_thread,
-                args=(
-                    process,
-                    heartbeat,
-                ),
-            )
-            heartbeat_thread.daemon = True
-            heartbeat_thread.start()
+
+        heartbeat_thread(process, heartbeat)
 
         if encoding is False:
             output_stdout = output_stderr = b""
@@ -942,9 +1021,23 @@ def command_runner(
 
         # Python >= 3.3 has SubProcessError(TimeoutExpired) class
         # Python >= 3.6 has encoding & error arguments
+        # Python >= 3.7 has creationflags arguments for process priority under windows
         # universal_newlines=True makes netstat command fail under windows
         # timeout does not work under Python 2.7 with subprocess32 < 3.5
         # decoder may be cp437 or unicode_escape for dos commands or utf-8 for powershell
+
+        if priority:
+            process_prio = _validate_process_priority(priority)
+            # Don't bother to make pylint go crazy on Windows missing os.nice()
+            # pylint: disable=E1101
+            if os_name == "nt" and sys.version_info >= (3, 7):
+                creationflags |= process_prio
+
+            # Actually we don't want preexec_fn since it's not thread safe, neither supported
+            # in subinterpreters. We'll use set_priority() once the process is spawned
+            # else:
+            #     kwargs["preexec_fn"] = lambda: os.nice(process_prio)
+
         # Disabling pylint error for the same reason as above
         # pylint: disable=E1123
         if sys.version_info >= (3, 6):
@@ -976,18 +1069,24 @@ def command_runner(
                 **kwargs
             )
 
-        # Set process priority if given
-        if priority:
+        # Set process priority if not set earlier by creationflags
+        if priority and (
+            os_name != "nt" or (sys.version_info < (3, 7) and os_name == "nt")
+        ):
             try:
                 try:
                     set_priority(process.pid, priority)
                 except psutil.AccessDenied as exc:
                     logger.warning(
-                        "Cannot set process priority {}. Access denied.".format(exc)
+                        "Cannot set process priority {}: {}. Access denied.".format(
+                            priority, exc
+                        )
                     )
                     logger.debug("Trace:", exc_info=True)
                 except Exception as exc:
-                    logger.warning("Cannot set process priority: {}".format(exc))
+                    logger.warning(
+                        "Cannot set process priority {}: {}".format(priority, exc)
+                    )
                     logger.debug("Trace:", exc_info=True)
             except NameError:
                 logger.warning(
@@ -1001,13 +1100,15 @@ def command_runner(
                     set_io_priority(process.pid, io_priority)
                 except psutil.AccessDenied as exc:
                     logger.warning(
-                        "Cannot set io priority for process {}: access denied.".format(
-                            exc
+                        "Cannot set io priority {}: {} Access denied.".format(
+                            io_priority, exc
                         )
                     )
                     logger.debug("Trace:", exc_info=True)
                 except Exception as exc:
-                    logger.warning("Cannot set io priority: {}".format(exc))
+                    logger.warning(
+                        "Cannot set io priority {}: {}".format(io_priority, exc)
+                    )
                     logger.debug("Trace:", exc_info=True)
                     raise
             except NameError:
@@ -1052,9 +1153,7 @@ def command_runner(
                 _stdout.write(output_stderr.encode(encoding, errors=errors))
 
         logger.debug(
-            'Command "{}" returned with exit code "{}". Command output was:\n{}'.format(
-                command, exit_code, to_encoding(output_stdout, error_encoding, errors)
-            )
+            'Command "{}" returned with exit code "{}"'.format(command, exit_code)
         )
     except subprocess.CalledProcessError as exc:
         exit_code = exc.returncode
@@ -1072,13 +1171,12 @@ def command_runner(
 
         if not silent:
             logger_fn(
-                'Command "{}" failed with{} exit code "{}". Command output was:'.format(
+                'Command "{}" failed with{} exit code "{}"'.format(
                     command,
                     valid_exit_codes_msg,
                     exc.returncode,
                 )
             )
-            logger_fn(output_stdout)
     except FileNotFoundError as exc:
         message = 'Command "{}" failed, file not found: {}'.format(
             command, to_encoding(exc.__str__(), error_encoding, errors)
@@ -1144,14 +1242,6 @@ def command_runner(
         if stderr_destination == "file":
             _stderr.close()
 
-    stdout_output = to_encoding(output_stdout, error_encoding, errors)
-    if stdout_output:
-        logger.debug("STDOUT: " + stdout_output)
-    if stderr_destination not in ["stdout", None]:
-        stderr_output = to_encoding(output_stderr, error_encoding, errors)
-        if stderr_output:
-            logger.debug("STDERR: " + stderr_output)
-
     # Make sure we send a simple queue end before leaving to make sure any queue read process will stop regardless
     # of command_runner state (useful when launching with queue and method poller which isn't supposed to write queues)
     if not no_close_queues:
@@ -1174,6 +1264,21 @@ def command_runner(
     ):
         output_stderr = None
 
+    stdout_output = to_encoding(output_stdout, error_encoding, errors)
+    if stdout_output:
+        logger.debug("STDOUT: " + stdout_output)
+    if stderr_destination not in ["stdout", None]:
+        stderr_output = to_encoding(output_stderr, error_encoding, errors)
+        if stderr_output and not silent:
+            if exit_code == 0 or (
+                valid_exit_codes
+                and valid_exit_codes is True
+                or exit_code in valid_exit_codes
+            ):
+                logger.debug("STDERR: " + stderr_output)
+            else:
+                logger.error("STDERR: " + stderr_output)
+
     if on_exit:
         logger.debug("Running on_exit callable.")
         on_exit()
@@ -1202,7 +1307,7 @@ def deferred_command(command, defer_time=300):
     seconds after it finished
     """
     # Use ping as a standard timer in shell since it's present on virtually *any* system
-    if os.name == "nt":
+    if os_name == "nt":
         deferrer = "ping 127.0.0.1 -n {} > NUL & ".format(defer_time)
     else:
         deferrer = "sleep {} && ".format(defer_time)
